#ifndef __UTILITIES_H__
#define __UTILITIES_H__

#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#ifndef uint
#define uint unsigned int
#endif

#define __FILENAME__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)

// Loggers
#define __FTD_LOG(_file, _fmt, ...)         fprintf(_file, "[%s:%d] " _fmt "\n", __FILENAME__, __LINE__, ##__VA_ARGS__)
#define INFO(_fmt, ...)                     __FTD_LOG(stdout, "[inf]: "_fmt, ##__VA_ARGS__)
#define WARNING(_fmt, ...)                  __FTD_LOG(stdout, "[war]: " _fmt, ##__VA_ARGS__)
#define ERROR(_fmt, ...)                    __FTD_LOG(stderr, "[err]: " _fmt, ##__VA_ARGS__)

// Catchers
#define __BASE_CATCHER(_expr, _fmt, extra, ...)                             \
{                                                                           \
    int _ret = _expr;                                                       \
    if(_ret < 0)                                                            \
        WARNING(_fmt " %s", ##__VA_ARGS__, extra);                          \
}
#define __BASE_CATCHER_CRITICAL(_expr, _fmt, extra, ...)                    \
{                                                                           \
    int _ret = _expr;                                                       \
    if(_ret < 0)                                                            \
    {                                                                       \
        ERROR(_fmt " %s", ##__VA_ARGS__, extra);                            \
        exit(_ret);                                                         \
    }                                                                       \
}

// Generic catchers
#define CATCHER(_expr, _fmt, ...)                                           \
    __BASE_CATCHER(_expr, _fmt, "", ##__VA_ARGS__)
#define CATCHER_CRITICAL(_expr, _fmt, ...)                                  \
    __BASE_CATCHER_CRITICAL(_expr, _fmt, "", ##__VA_ARGS__)

// stderror catchers
#define STD_CATCHER(_expr, _fmt, ...)                                       \
    __BASE_CATCHER(_expr, _fmt ":", strerror(errno), ##__VA_ARGS__)
#define STD_CATCHER_CRITICAL(_expr, _fmt, ...)                              \
    __BASE_CATCHER_CRITICAL(_expr, _fmt ":", strerror(errno), ##__VA_ARGS__)

#define VALIDATE(_expr, _fmt, ...) CATCHER_CRITICAL((_expr) ? 0 : -1, _fmt, ##__VA_ARGS__);
#define VALIDATE_STD(_expr, _fmt, ...) STD_CATCHER_CRITICAL((_expr) ? 0 : -1, _fmt, ##__VA_ARGS__);

extern int errno;

// Misc
#define GET_LEN(_arr)           (sizeof(_arr)/sizeof(_arr[0]))
#define SLEEP_FOR_US(_us)                                                   \
{                                                                           \
    struct timespec req = {                                                 \
        .tv_nsec = (long)(_us)*1000                                         \
    };                                                                      \
    while(nanosleep(&req, &req) == -1 && errno == EINTR);                   \
}
#define UNUSED(x) (void)(x)

// Mutex helpers
#define REGISTER_MUTEX_VARIABLE(_type, _name)                               \
    static pthread_mutex_t _name##_mutex;                                   \
    static _type _name;                                                     \
    static inline void lock_##_name()                                       \
    { pthread_mutex_lock(&_name##_mutex); }                                 \
    static inline void unlock_##_name()                                     \
    { pthread_mutex_unlock(&_name##_mutex); }                               \
    static void set_##_name(_type val)                                      \
    {                                                                       \
        lock_##_name();                                                     \
        _name = val;                                                        \
        unlock_##_name();                                                   \
    }                                                                       \
    static _type get_##_name()                                              \
    {                                                                       \
        _type _v;                                                           \
        lock_##_name();                                                     \
        _v = _name;                                                         \
        unlock_##_name();                                                   \
        return _v;                                                          \
    }

#define GENERIC_VECTOR(NAME, T)                                             \
    typedef struct NAME {                                                   \
        T *data;                                                            \
        size_t used;                                                        \
        size_t reserved;                                                    \
    } NAME;                                                                 \
    static inline void NAME##_init(NAME *v) {                               \
        memset(v, 0, sizeof(NAME));                                         \
    }                                                                       \
    static int NAME##_reserve(NAME *v, size_t n) {                          \
        v->reserved += n;                                                   \
        v->data = (T *)reallocarray(v->data, v->reserved, sizeof(T));       \
        return v->data != NULL ? 0 : -1;                                    \
    }                                                                       \
    static int NAME##_shrink(NAME *v) {                                     \
        v->reserved = v->used;                                              \
        v->data = (T *)reallocarray(v->data, v->reserved, sizeof(T));       \
        return v->data != NULL ? 0 : -1;                                    \
    }                                                                       \
    static void NAME##_free(NAME *v) {                                      \
        if(v->data == NULL) return;                                         \
        free(v->data);                                                      \
        memset(v, 0, sizeof(T));                                            \
    }                                                                       \
    static inline T* NAME##_get_data(NAME *v) {                             \
        return v->data;                                                     \
    }                                                                       \
    static inline size_t NAME##_get_len(NAME *v) {                          \
        return v->used/sizeof(T);                                           \
    }

#define MEMSET_ZERO(_v)                  memset(&_v, 0, sizeof(_v))

#define CHECK_MASK(_value, _mask) (((_value) & (_mask)) == 0)

#define MAX(a, b) ((a) >= (b) ? (a): (b))
#define MIN(a, b) ((a) <= (b) ? (a): (b))

#endif
